@@ -2,15 +2,21 @@ package Devel::GlobalPhase;
 use strict;
 use warnings;
 
-our $VERSION = '0.002004';
+our $VERSION = '0.003000';
 $VERSION = eval $VERSION;
 
 use base 'Exporter';
 
 our @EXPORT = qw(global_phase);
 
+BEGIN {
+  *_CALLER_CAN_SEGFAULT = ("$]" >= 5.008009 && "$]" < 5.018000) ? sub(){1} : sub(){0};
+  *_NATIVE_GLOBAL_PHASE = "$]" >= 5.014000 ? sub(){1} : sub(){0};
+}
+
 sub global_phase ();
 sub tie_global_phase;
+sub _refresh_END ();
 
 sub import {
   my $class = shift;
@@ -24,13 +30,14 @@ sub import {
   if (@imports || !$var) {
     Exporter::export_to_level($class, 1, @imports);
   }
+  _refresh_END;
   if ($var) {
     tie_global_phase;
   }
 }
 
 BEGIN {
-  if (defined ${^GLOBAL_PHASE}) {
+  if (_NATIVE_GLOBAL_PHASE) {
     eval <<'END_CODE' or die $@;
 
 sub global_phase () {
@@ -39,6 +46,8 @@ sub global_phase () {
 
 sub tie_global_phase { 1 }
 
+sub _refresh_END () { 1 }
+
 1;
 
 END_CODE
@@ -55,7 +64,7 @@ if (B::main_start()->isa('B::NULL')) {
 
     CHECK { $global_phase = 'CHECK' }
     # try to install an END block as late as possible so it will run first.
-    INIT  { eval q( END { $global_phase = 'END' } ) }
+    INIT  { my $capture = $global_phase; eval q( END { $global_phase = 'END' } ) }
     # INIT is FIFO so we can force our sub to be first
     unshift @{ B::init_av()->object_2svref }, sub { $global_phase = 'INIT' };
 
@@ -69,6 +78,11 @@ else {
 }
 END { $global_phase = 'END' }
 
+sub _refresh_END () {
+  my $capture = $global_phase;
+  eval q[ END { $global_phase = 'END' } ];
+}
+
 sub global_phase () {
   if ($global_phase eq 'DESTRUCT') {
     # no need for extra checks at this point
@@ -84,37 +98,44 @@ sub global_phase () {
     $global_phase = 'DESTRUCT';
   }
   elsif ($global_phase eq 'INIT' && B::main_cv()->DEPTH > 0) {
+    _refresh_END;
     $global_phase = 'RUN';
   }
 
-  if ($global_phase eq 'RUN' && $^S) {
+  # this is slow and can segfault, so skip it
+  if (!_CALLER_CAN_SEGFAULT && $global_phase eq 'RUN' && $^S) {
     # END blocks are FILO so we can't install one to run first.
     # only way to detect END reliably seems to be by using caller.
     # I hate this but it seems to be the best available option.
     # The top two frames will be an eval and the END block.
     my $i = 0;
-    $i++ while CORE::caller($i + 1);
+    $i++ while defined CORE::caller($i + 1);
     if ($i < 1) {
       # there should always be the sub call and an eval frame ($^S is true).
       # this will only happen if we're in END, but the outer frames are broken.
       $global_phase = 'END';
     }
     elsif ($i > 1) {
-      # If we're ENDing due to an exit or die in a sub generated in an eval,
-      # these caller calls can cause a segfault.  I can't find a way to detect
-      # this.
-      my @top = CORE::caller($i);
-      my @next = CORE::caller($i - 1);
-      if (
-        $top[3] eq '(eval)'
-        && $next[3] =~ /::END$/
-        && $top[2] == $next[2]
-        && $top[1] eq $next[1]
-        && $top[0] eq 'main'
-        && $next[0] eq 'main'
-      ) {
+      my $top = CORE::caller($i);
+      my $next = CORE::caller($i - 1);
+      if (!$top || !$next) {
         $global_phase = 'END';
       }
+      elsif ($top eq 'main' && $next eq 'main') {
+        # If we're ENDing due to an exit or die in a sub generated in an eval,
+        # these caller calls can cause a segfault.  I can't find a way to detect
+        # this.
+        my @top = CORE::caller($i);
+        my @next = CORE::caller($i - 1);
+        if (
+          $top[3] eq '(eval)'
+          && $next[3] =~ /::END$/
+          && $top[2] == $next[2]
+          && $top[1] eq $next[1]
+        ) {
+          $global_phase = 'END';
+        }
+      }
     }
   }
 
@@ -126,8 +147,14 @@ sub global_phase () {
     Devel::GlobalPhase::_Tie;
 
   sub TIESCALAR { bless \(my $s), $_[0]; }
-  sub STORE { die sprintf "Modification of a read-only value attempted at %s line %s.\n", (caller(0))[1,2]; }
-  *FETCH = \&Devel::GlobalPhase::global_phase;
+  sub STORE {
+    die sprintf "Modification of a read-only value attempted at %s line %s.\n", (caller(0))[1,2];
+  }
+  sub FETCH {
+    return undef
+      if caller eq 'Devel::GlobalDestruction';
+    Devel::GlobalPhase::global_phase;
+  }
   sub DESTROY {
     my $tied = tied ${^GLOBAL_PHASE};
     if ($tied && $tied == $_[0]) {
@@ -192,9 +219,22 @@ built in variable from newer perls.
 
 =head1 BUGS
 
+=over 4
+
+=item *
+
 There are tricks that can be played with L<B> or XS that would fool this
 module for the INIT and END phase.
 
+=item *
+
+During an C<END {}> block created at runtime after this module is loaded, the
+phase may be reported as C<RUN>.  While this could be made more accurate, it
+would slow down the module significantly during the RUN phase, and has the
+potential to segfault perl.
+
+=back
+
 =head1 AUTHOR
 
 haarg - Graham Knop (cpan:HAARG) <haarg@haarg.org>
