@@ -12,8 +12,6 @@ for (@ARGV) {
 my $code = "END {\n" . ( "(sub {\n" x $layers ) . <<'END_CODE' . ("})->()\n" x $layers) . "}\n1";
   my $phase = global_phase;
   print "$phase\n";
-  $? = 1
-    if $phase ne 'END';
 END_CODE
 eval $code or die $@;
 
